import os
import re
import copy
from pathlib import Path
from collections import defaultdict
from typing import Union, List, Dict
from functools import cached_property

from sphinx.application import Sphinx
from sphinx.errors import ExtensionError

from sphinx_readme.utils import get_conf_val, set_conf_val, logger, replace_only_directives, remove_raw_directives
from sphinx_readme.config import get_repo_dir, get_repo_host, get_blob_url, get_repo_url, get_linkcode_url, get_linkcode_resolve


class READMEConfig:

    REFERENCE_MAPPING = {
        "replace": None,
        "target": None,
    }

    def __init__(self, app: Sphinx):
        self.logger = logger
        self.src_dir = Path(app.srcdir)
        self.repo_dir = get_repo_dir()
        self.out_dir = get_conf_val(app, 'readme_out_dir')
        self.src_files = get_conf_val(app, 'readme_src_files', [])
        self.html_context = get_conf_val(app, "html_context")
        self.html_baseurl = get_conf_val(app, "html_baseurl", "").rstrip("/")
        self.docs_url_type = get_conf_val(app, 'readme_docs_url_type')
        self.replace_attrs = get_conf_val(app, 'readme_replace_attrs')
        self.inline_markup = get_conf_val(app, 'readme_inline_markup')
        self.raw_directive = get_conf_val(app, 'readme_raw_directive')
        self.rubric_heading = get_conf_val(app, 'readme_rubric_heading')
        self.admonition_icons = get_conf_val(app, 'readme_admonition_icons')
        self.include_directive = get_conf_val(app, 'readme_include_directive')
        self.default_admonition_icon = get_conf_val(app, 'readme_default_admonition_icon')

        self.repo_blob = get_conf_val(app, "readme_blob")
        self.repo_url = self.get_repo_url()
        self.blob_url = get_blob_url(
            repo_url=self.repo_url,
            blob=self.repo_blob
        )
        self.docs_url = self.get_docs_url()
        self.ref_map = self.get_ref_map()
        self.source_files = self.read_source_files()

        if self.docs_url_type == "code":
            self.setup_linkcode_resolve(app)

    def get_repo_url(self) -> str:
        """Generates the repository URL from the :external+sphinx:confval:`html_context` dict

        :raises ExtensionError: if ``html_context`` is not set or missing values
        """
        if not self.html_context:
            raise ExtensionError(
                "``sphinx_readme``: conf.py value "
                "must be set for ``html_context``"
            )
        return get_repo_url(self.html_context)

    def get_docs_url(self) -> str:
        """Returns the base URL of the documentation source to use
        when resolving :mod:`~.sphinx.ext.autodoc` cross-references

         If :attr:`docs_url_type` is

         * ``"code"``: uses the :attr:`blob_url`
         * ``"html"``: uses the :confval:`html_baseurl`

        :raises ExtensionError: if ``html_baseurl`` is missing
        """
        if self.docs_url_type == "html":
            if self.html_baseurl:
                docs_url = self.html_baseurl
            else:
                raise ExtensionError(
                    "``sphinx_readme``: conf.py value "
                    "must be set for ``html_baseurl``"
                )
        else:  # ``docs_url_type`` is "code"
            docs_url = self.blob_url

        return docs_url.rstrip("/")

    def setup_linkcode_resolve(self, app: Sphinx) -> None:
        linkcode_func = get_conf_val(app, "linkcode_resolve")

        if not callable(linkcode_func):
            self.logger.info(
                "``sphinx_readme:`` using default ``linkcode_resolve``"
            )
            # Get the template for linking to source code
            linkcode_url = get_linkcode_url(self.blob_url)
            linkcode_func = get_linkcode_resolve(linkcode_url)

        set_conf_val(app, 'linkcode_resolve', linkcode_func)
        app.setup_extension("sphinx.ext.linkcode")

    def get_ref_map(self) -> Dict:
        refs = defaultdict(_map_entry)
        refs['ref'] = []
        refs['doc'] = []
        return refs

    def read_source_files(self) -> Dict[str, str]:
        sources = {
            src_file: self.read_rst(src_file)
            for src_file in self.src_files
        }
        return sources

    def read_rst(self, rst_file: Union[str, Path], replace_only: bool = False) -> str:
        with open(rst_file, 'r', encoding='utf-8') as f:
            rst = f.read()

        if replace_only:
            rst = replace_only_directives(rst)

        include_pattern = r"^\.\. include:: ([./]*?[\w/-]+\.rst)"

        if self.include_directive:
            # Find all included files
            included = re.findall(
                pattern=include_pattern,
                string=rst,
                flags=re.M
            )
            for include in included:
                # Determine abs path of included file
                if include.startswith("/"):
                    # These paths are relative to source dir
                    file = Path(f"{self.src_dir}{include}").resolve()
                else:
                    # These paths are relative to rst_file dir
                    file = (Path(rst_file).parent / Path(include)).resolve()

                # Sub in the file content
                rst = re.sub(
                    pattern=rf".. include:: {include}",
                    repl=self.read_rst(file, replace_only).replace(r'\n', r'\\n'),
                    string=rst
                )
        else:
            # Remove all include directives from the text
            rst = re.sub(include_pattern, '', rst, flags=re.M)

        if self.raw_directive is False:
            rst = remove_raw_directives(rst)

        return rst

    @property
    def src_files(self):
        return self._src_files

    @src_files.setter
    def src_files(self, src_files: Union[str, List[str]]):
        if isinstance(src_files, str):
            src_files = [src_files]

        self._src_files = [  # Absolute paths of files; files should be relative to source directory
            str((Path(self.src_dir) / Path(src_file)).resolve())
            for src_file in src_files
        ]

    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, out: str):
        out_dir = Path(out)

        if not out_dir.is_absolute():
            out_dir = (self.src_dir / out_dir).resolve()

        if not out_dir.exists():
            out_dir.mkdir()

        self._out_dir = out_dir

    @property
    def docs_url_type(self):
        return self._docs_url_type

    @docs_url_type.setter
    def docs_url_type(self, docs_url_type: str):
        if docs_url_type not in ('html', 'code'):
            raise ExtensionError(
                "``sphinx_readme``: ``readme_docs_url_type`` value must be ``code`` or ``html``"
            )
        if docs_url_type == "code" and not self.html_baseurl:
            raise ExtensionError(  # HTML url is needed for non-source code xrefs
                "``sphinx_readme``: conf.py value missing for ``html_baseurl``"
            )
        self._docs_url_type = docs_url_type

    @cached_property
    def repo_host(self):
        return get_repo_host(self.repo_url)

    @cached_property
    def image_baseurl(self):
        if self.repo_host == "github":
            # Ex. https://raw.githubusercontent.com/TDKorn/sphinx-readme/main
            return self.blob_url.replace("github.com", "raw.githubusercontent.com").replace('blob/', '')

        elif self.repo_host == "gitlab":
            # Ex. https://gitlab.com/TDKorn/sphinx-readme/raw/main
            return self.blob_url.replace("/blob/", "/raw/")

        else:
            # Ex. https://bitbucket.org/TDKorn/sphinx-readme/raw/main
            return self.blob_url.replace("/src/", "/raw/")


    @cached_property
    def icon_map(self):
        types = ("attention", "caution", "danger", "error", "hint", "important", "note", "tip", "warning", "default")
        icons = ("🔔️", "⚠️", "☢️", '⛔', "🧠", "📢", "📝", "💡", "🚩", self.default_admonition_icon)
        icon_map = dict(zip(types, icons))

        # Update/add custom admonition icons from conf.py
        if self.admonition_icons:
            if isinstance(self.admonition_icons, Dict):
                icon_map.update(self.admonition_icons)

        return icon_map

    @property
    def admonition_template(self):
        if self.raw_directive is True:
            return r'''
.. raw:: html

   <table>
       <tr align="left">
           <th>

{icon} {title}

.. raw:: html

   </th>
   <tr><td>

{text}

.. raw:: html

   </td></tr>
   </table>
'''
        else:
            return r'''
.. csv-table::
   :header: {icon} {title}

   "\1"
'''


def _map_entry():
    return copy.deepcopy(READMEConfig.REFERENCE_MAPPING)



import os
import re
import sys
import inspect
import subprocess
import pkg_resources
from pathlib import Path
from typing import Dict, Optional, Callable
from sphinx.errors import ExtensionError
from sphinx_readme.utils import logger


def get_linkcode_url(blob_url: Optional[str] = None,
                     repo_url: Optional[str] = None,
                     context: Optional[Dict] = None,
                     blob: Optional[str] = None) -> str:
    """Get the template URL for linking to highlighted GitHub source code

    Formatted into the final link by ``linkcode_resolve()``
    """
    if blob_url is None:
        if repo_url is None:
            if context is None:
                raise ExtensionError(
                    "``sphinx_readme:`` config value ``html_context`` is missing")
            else:
                repo_url = get_repo_url(context)
        blob_url = get_blob_url(repo_url, blob, context)

    if 'bitbucket' in blob_url:
        return blob_url + "/{filepath}#lines-{linestart}:{linestop}"
    else:
        return blob_url + "/{filepath}#L{linestart}-L{linestop}"


def get_repo_url(context: Dict):
    for host in ('github', 'gitlab', 'bitbucket'):
        user = context.get(f"{host}_user")
        repo = context.get(f"{host}_repo")

        if not all((user, repo)):
            continue

        if host == "bitbucket":
            tld = "org"
        else:
            tld = "com"

        return f"https://{host}.{tld}/{user}/{repo}"

    raise ExtensionError(
        "``sphinx_readme``: unable to determine repo url")


def get_blob_url(repo_url: str, blob: Optional[str] = None, context: Optional[Dict] = None) -> str:
    """Generate the url for a specific blob of a repository

    If ``blob`` and ``context`` are not provided, the most recent commit hash will be used
    """
    if context:
        host = get_repo_host(repo_url)
        blob = context.get(f"{host}_version")

    if blob is not None:
        # Use blob from kwarg/html_context
        blob = get_linkcode_revision(blob)
    else:
        # Use hash of the most recent commit
        blob = get_linkcode_revision('head')

    if "bitbucket" in repo_url:
        return repo_url.strip('/') + f"/src/{blob}"
    else:
        return repo_url.strip("/") + f"/blob/{blob}"


def get_repo_host(url: str):
    if match := re.match(r"https?://(\w+)\.(?:com|org)", url):
        return match.group(1)
    return None


def get_linkcode_revision(blob: str) -> str:
    """Get the blob to link to

    .. note::

       The value of ``blob`` can be any of ``"head"``, ``"last_tag"``, or ``"{blob}"``

       * ``head`` (default): links to the most recent commit hash; if this commit is tagged, uses the tag instead
       * ``last_tag``: links to the most recently tagged commit; if no tags exist, uses ``head``
       * ``blob``: links to any blob you want, for example ``"master"`` or ``"v2.0.1"``
    """
    if blob == "head":
        return get_head()
    if blob == 'last_tag':
        return get_last_tag()
    # Link to the branch/tree/blob you provided, ex. "master"
    return blob


def get_head(errors: bool = False) -> Optional[str]:
    """Gets the most recent commit hash or tag

    :raises subprocess.CalledProcessError: if the commit can't be found and ``errors`` is ``True``
    """
    cmd = "git log -n1 --pretty=%H"
    try:
        # get most recent commit hash
        head = subprocess.check_output(cmd.split()).strip().decode('utf-8')

        # if head is a tag, use tag as reference
        cmd = "git describe --exact-match --tags " + head
        try:
            tag = subprocess.check_output(cmd.split(" ")).strip().decode('utf-8')
            return tag

        except subprocess.CalledProcessError:
            return head

    except subprocess.CalledProcessError as e:
        if errors:
            raise e
        else:
            return logger.error("Failed to get head")  # so no head?


def get_last_tag() -> str:
    """Get the most recent commit tag

    :raises RuntimeError: if there are no tagged commits
    """
    try:
        cmd = "git describe --tags --abbrev=0"
        return subprocess.check_output(cmd.split(" ")).strip().decode('utf-8')

    except subprocess.CalledProcessError as e:
        raise RuntimeError("No tags exist for the repo") from e


def get_repo_dir() -> Path:
    """Get the root directory of the repository

    :raises RuntimeError: if the directory can't be determined
    """
    try:
        cmd = "git rev-parse --show-toplevel"
        repo_dir = Path(subprocess.check_output(cmd.split(" ")).strip().decode('utf-8'))

    except subprocess.CalledProcessError as e:
        raise RuntimeError("Unable to determine the repository directory") from e

    # For ReadTheDocs, repo is cloned to /path/to/<repo_dir>/checkouts/<version>/
    if repo_dir.parent == "checkouts":
        return repo_dir.parent.parent
    else:
        return repo_dir


def get_linkcode_resolve(linkcode_url: str) -> Callable:
    """Defines and returns a ``linkcode_resolve`` function for your package

    Used by default if ``linkcode_resolve`` isn't defined in ``conf.py``
    """

    repo_dir = get_repo_dir()
    pkg = pkg_resources.require(repo_dir.name)[0]
    top_level = pkg.get_metadata('top_level.txt').strip()

    def linkcode_resolve(domain, info):
        """Returns a link to the source code on GitHub, with appropriate lines highlighted

        :By:
            Adam Korn (https://github.com/tdkorn)
        :Adapted From:
            nlgranger/SeqTools (https://github.com/nlgranger/seqtools/blob/master/docs/conf.py)
        """
        if domain != 'py' or not info['module']:
            return None

        modname = info['module']
        fullname = info['fullname']

        submod = sys.modules.get(modname)
        if submod is None:
            return None

        obj = submod
        for part in fullname.split('.'):
            try:
                obj = getattr(obj, part)
            except AttributeError:
                return None

        try:
            filepath = os.path.relpath(inspect.getsourcefile(obj), repo_dir)
            if filepath is None:
                return
        except Exception:
            return None

        try:
            source, lineno = inspect.getsourcelines(obj)
        except OSError:
            return None
        else:
            linestart, linestop = lineno, lineno + len(source) - 1

        # Fix links with "../../../" or "..\\..\\..\\"
        filepath = '/'.join(filepath[filepath.find(top_level):].split('\\'))

        # Example: https://github.com/TDKorn/my-magento/blob/docs/magento/models/model.py#L28-L59
        final_link = linkcode_url.format(
            filepath=filepath,
            linestart=linestart,
            linestop=linestop
        )
        logger.debug(f"Final Link for {fullname}: {final_link}")
        return final_link

    return linkcode_resolve



import re
from copy import copy
from pathlib import Path
from collections import defaultdict
from typing import Dict, List, Tuple, Set

from docutils import nodes
from docutils.nodes import Node, document

from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.testing import restructuredtext
from sphinx.transforms import SphinxTransformer
from sphinx.environment import BuildEnvironment

from sphinx_readme.config import READMEConfig
from sphinx_readme.utils import get_all_variants, escape_rst, format_rst


class READMEParser:

    def __init__(self, app: Sphinx):
        self.config = READMEConfig(app)
        self.logger = self.config.logger
        self.ref_map = self.config.ref_map
        self.sources = self.config.source_files
        self.toctrees = defaultdict(list)
        self.admonitions = {}
        self.titles = {}

    def parse(self, app: Sphinx, doctree: document, docname: str):
        # If a source has ``only`` directives, its doctree will have missing/extra content
        # Replace the ``only`` directives, then generate a new doctree if needed
        doctree = self.get_doctree(app, doctree, docname)

        inline_nodes = list(doctree.findall(nodes.inline))
        literal_nodes = list(doctree.findall(nodes.literal))

        self.parse_autodoc_nodes(literal_nodes, docname)

        if self.config.docs_url_type == "code":
            self.parse_linkcode_nodes(inline_nodes)

        for node in list(doctree.findall(addnodes.toctree)):
            self.parse_toctree(node, doctree)

        if doctree.get('source') in self.sources:
            self.parse_xref_nodes(inline_nodes)
            self.parse_admonitions(doctree)

    def get_doctree(self, app: Sphinx, doctree: document, docname: str) -> document:
        # Return original doctree if file is not a readme source
        if (src := doctree.get('source')) not in self.sources:
            return doctree

        # Parse ``only`` directives to get true source rst of README version
        parsed_rst = self.config.read_rst(src, replace_only=True)
        raw_rst = self.sources[src]

        # Return original doctree if file had no ``only`` directives
        if parsed_rst == raw_rst:
            return doctree

        self.sources[src] = parsed_rst

        # Use temp docname to avoid duplicate warnings
        docname = docname + "_readme"

        # Generate new doctree from parsed rst using Sphinx application
        doctree = restructuredtext.parse(app, parsed_rst, docname)

        # Resolve references in the doctree
        try:
            backup = copy(app.env.temp_data)
            app.env.temp_data['docname'] = docname

            transformer = SphinxTransformer(doctree)
            transformer.set_environment(app.env)
            transformer.add_transforms(app.registry.get_post_transforms())
            transformer.apply_transforms()

        finally:
            app.env.temp_data = backup

        # Replace temp source with actual source
        doctree['source'] = src
        return doctree

    def parse_xref_nodes(self, inline_nodes: List[nodes.inline]):
        for node in inline_nodes:
            if 'doc' in node['classes']:
                self.ref_map['doc'].append(self._parse_xref(node))

            elif 'std-ref' in node['classes']:
                self.ref_map['ref'].append(self._parse_xref(node))

    def _parse_xref(self, node: nodes.inline) -> Dict:
        """Helper function to parse target info of a ``:ref:`` or ``:doc:`` xref"""
        return {
            'text': node.children[0].astext(),
            "refuri": self.config.html_baseurl + "/" + node.parent.get('refuri', '')
        }

    def parse_autodoc_nodes(self, literal_nodes: List[nodes.literal], docname: str):
        for node in literal_nodes:
            if 'py' not in node['classes']:
                continue

            if not isinstance(node.parent, nodes.reference):
                continue

            if node.parent.get('internal') is False:
                # External links are xrefs from intersphinx
                self.parse_intersphinx_node(node)

            elif 'py-mod' in node['classes']:
                # Parse :mod: xrefs regardless of docs_url_type
                self.parse_module_node(node, docname)

            elif self.config.docs_url_type == "html":
                # Only parse remaining py xrefs if linking to html
                self.parse_autodoc_node(node, docname)

    def parse_intersphinx_node(self, node: nodes.literal):
        pattern = r":(mod|class|meth|func|attr):`~?\.?[.\w]+`"
        match = re.match(pattern, node.rawsource)
        target = node.parent.get('refuri')

        if not all((match, target)):
            return

        is_callable = match.group(1) in ("meth", "func")
        qualified_name = target.split("#")[-1].split("-")[-1]
        self.add_variants(qualified_name, target, is_callable)

    def parse_module_node(self, node: Node, docname: str):
        qualified_name = node.parent.get("reftitle", "")

        if self.config.docs_url_type == 'code':
            # Ex. sphinx_readme.parser -> sphinx_readme/parser.py
            refuri = qualified_name.replace('.', '/') + '.py'
        else:
            refuri = self._parse_refuri(node, docname)

        if not all((qualified_name, refuri)):
            return

        target = f"{self.config.docs_url}/{refuri}"
        self.add_variants(qualified_name, target)

    def parse_autodoc_node(self, node: nodes.literal, docname: str):
        qualified_name = node.parent.get("reftitle")
        refuri = self._parse_refuri(node, docname)

        if not all((refuri, qualified_name)):
            return

        is_callable = bool(re.match(r":(meth|func):", node.rawsource))
        target = f"{self.config.docs_url}/{refuri}"
        self.add_variants(qualified_name, target, is_callable)

    def _parse_refuri(self, node: Node, docname: str):
        if 'refuri' in node.parent:
            # Ex. ../parser.html#sphinx_readme.parser.READMEParser
            return node.parent["refuri"].lstrip("./")

        elif 'refid' in node.parent:
            # Ex. sphinx_readme.parser.READMEParser
            return f"{docname}.html#{node.parent['refid']}"

    def parse_linkcode_nodes(self, inline_nodes: List[nodes.inline]):
        for node in inline_nodes:
            if 'viewcode-link' in node['classes'] or 'linkcode-link' in node['classes']:
                if node.parent.get('internal') is False:
                    # Only parse links to external source code
                    self.parse_linkcode_node(node)

    def parse_linkcode_node(self, node: nodes.inline):
        grandparent = node.parent.parent
        is_callable = grandparent.get("_toc_name", "").endswith("()")

        try:
            qualified_name = grandparent.get("ids")[0]
        except IndexError:
            qualified_name = grandparent.get("module", "") + grandparent.get("fullname", "")

        target = node.parent.get("refuri")
        self.add_variants(qualified_name, target, is_callable)

    def add_variants(self, qualified_name, target, is_callable: bool = False):
        short_ref = qualified_name.split('.')[-1]
        variants = get_all_variants(qualified_name)

        for variant in variants:
            if variant in self.ref_map:
                continue

            if variant.startswith("~"):
                replace = short_ref
            else:
                replace = variant.lstrip('.')

            if is_callable:
                replace += "()"

            if self.config.inline_markup:
                replace = f"``{replace}``"

            self.ref_map[variant].update({
                'target': target,
                'replace': replace
            })

    def parse_titles(self, env: BuildEnvironment):
        for fname, title_node in env.titles.items():
            parts = []

            for child in title_node.children:
                text = child.astext()
                if isinstance(child, nodes.literal):
                    parts.append(f"``{text}``")
                else:
                    parts.append(text)

            self.titles[fname] = ' '.join(parts)

    def parse_toctree(self, toctree: Node, doctree: addnodes.document):
        src = doctree.get('source')
        toc = {
            'caption': toctree.get('caption'),
            'entries': []
        }
        for _, entry in toctree.get('entries', []):
            toc['entries'].append({
                'entry': entry,
                'title': self.titles.get(entry),
            })
        self.toctrees[src].append(toc)

    def parse_admonitions(self, doctree: Node):
        admonitions = {'generic': [], 'specific': []}
        src = doctree.get('source')

        for admonition in list(doctree.findall(nodes.Admonition)):
            info = {
                'body': admonition.rawsource
            }
            if isinstance(admonition, nodes.admonition):
                # Generic Admonition (using admonition directive)
                info.update({
                    'class': admonition.get('classes')[0],
                    'title': admonition.children[0].rawsource
                })
                admonitions['generic'].append(info)
            else:
                # Specific Admonition (for example, .. note::)
                info.update({
                    'class': admonition.tagname,
                    'title': admonition.tagname.title(),
                })
                admonitions['specific'].append(info)

        self.admonitions[src] = admonitions

    def resolve(self):
        for src, rst in self.sources.items():
            # Replace everything using data from ``parse()``
            rst = self.replace_admonitions(src, rst)
            rst = self.replace_rst_images(src, rst)
            rst = self.replace_toctrees(src, rst)
            rst = self.replace_rst_rubrics(rst)

            for role in ('ref', 'doc'):
                rst = self.replace_cross_refs(rst, role)

            # Use ref_map to generate autodoc substitution definitions
            rst, autodoc_refs = self.replace_autodoc_refs(rst)
            header_vals = self.get_header_vals(autodoc_refs)

            # Write the final output
            rst_out = Path(self.config.out_dir, Path(src).name)

            with open(rst_out, 'w', encoding='utf-8') as f:
                f.write(
                    "\n".join(header_vals) + "\n\n" + rst)
            print(
                f'``sphinx_readme``: saved generated .rst file to {rst_out}')

    def replace_admonitions(self, rst_src: str, rst: str) -> str:
        admonitions = self.admonitions[rst_src]

        for _type in ('generic', 'specific'):
            for admonition in admonitions[_type]:
                if pattern := self.get_admonition_regex(admonition, _type):
                    icon = self.get_admonition_icon(admonition)
                    if not self.config.raw_directive:
                        rst = re.sub(
                            pattern=pattern,
                            repl=self.config.admonition_template.format(
                                title=admonition['title'],
                                icon=icon),
                            string=rst
                        )
                    else:
                        rst = re.sub(
                            pattern=pattern,
                            repl=self.config.admonition_template.format(
                                title=admonition['title'],
                                text=admonition['body'],
                                icon=icon),
                            string=rst
                        )
        return rst

    def replace_toctrees(self, rst_src: str, rst: str) -> str:
        if self.config.docs_url_type == 'html':
            base_url = self.config.docs_url
        else:
            base_url = self.config.html_baseurl

        pattern = r".. toctree::\n+?(?:\s+:\w+:\s*?\w*?\n\s+)*?(?:\s+\w+\n)+?(?=\n+\S+?)"
        toctrees = re.findall(pattern, rst)

        for toctree, info in zip(toctrees, self.toctrees[rst_src]):
            substitutions = []
            repl = ""

            if info['caption']:
                repl += f"**{info['caption']}**\n\n"

            for entry in info['entries']:
                # Replace each entry with a link to html docs
                target = f"{base_url}/{entry['entry']}.html"

                if "`" in entry['title']:
                    # Inline markup in links must be inserted with substitutions
                    sub = entry['title'].replace('`', '')
                    substitutions.extend([
                        f".. |{sub}| replace:: {entry['title']}",
                        f".. _{sub}: {target}"
                    ])
                    repl += f"* |{sub}|_\n"

                else:
                    # Replace with a normal link otherwise
                    repl += f"* `{entry['title']} <{target}>`_\n"

            if substitutions:
                repl += '\n\n' + '\n'.join(substitutions) + '\n\n'

            # Replace toctree directive with links and substitution defs
            rst = rst.replace(toctree, repl)

        return rst

    def replace_rst_images(self, rst_src: str, rst: str) -> str:
        """Replaces filepaths in ``image`` directives with repository links

        :Example:
            :rst:`.. image:: /_static/logo.png`

            would be replaced with

            :rst:`.. image:: https://github.com/tdkorn/sphinx-readme/blob/main/docs/source/_static/logo.png?raw=True`

        .. note:: Your repository will be used as the image source regardless of the
           value of :confval:`readme_docs_url_type`

        :param rst_src: absolute path of the rst file being parsed
        :param rst: the content of the rst file being parsed
        """
        src_dir = self.config.src_dir
        repo_dir = self.config.repo_dir
        rst_src_dir = Path(rst_src).parent
        relpath_to_src_dir = src_dir.relative_to(repo_dir)

        # Find the targets of all image directives
        img_pattern = r"\.\. image:: ([./\w-]+\.\w{3,4})"
        img_paths = re.findall(img_pattern, rst)

        for img_path in img_paths:
            if img_path.startswith("/"):
                # These paths are relative to source dir
                path_to_img = Path(f"{relpath_to_src_dir}{img_path}").as_posix()

            else:
                # These paths are relative to rst_file dir
                abs_img_path = (rst_src_dir / Path(img_path)).resolve()

                # Find path of image relative to the repo directory
                path_to_img = abs_img_path.relative_to(repo_dir).as_posix()

            # Sub that hoe in!!!
            rst = re.sub(
                pattern=rf"\.\. image:: {img_path}",
                repl=fr".. image:: {self.config.image_baseurl}/{path_to_img}",
                string=rst
            )
        return rst

    def replace_rst_rubrics(self, rst: str):
        heading_chars = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
        rubric_pattern = r'\.\. rubric:: (.+?)(?=\n)'

        if heading := self.config.rubric_heading:
            if heading in heading_chars:
                return re.sub(
                    pattern=rubric_pattern,
                    repl=lambda m: f"{m.group(1)}\n{heading * len(m.group(1))}",
                    string=rst
                )
        return re.sub(
            pattern=rubric_pattern,
            repl=lambda m: format_rst("bold", m.group(1)),
            string=rst
        )

    def replace_cross_refs(self, rst: str, ref_role: str) -> str:
        # Find all :ref_role:`ref_id` cross-refs
        cross_refs = re.findall(
            pattern=fr"(?:\s*?):{ref_role}:`([^`]+)`(?=\s*?)",
            string=rst
        )
        # Match these ids up with target data in the ref_map
        cross_ref_map = dict(zip(cross_refs, self.ref_map[ref_role]))

        # Replace cross-refs with `text <link>`_ format
        for ref_id, target in cross_ref_map.items():
            rst = re.sub(
                pattern=rf":{ref_role}:`{ref_id}`",
                repl=rf"`{target['text']} <{target['refuri']}>`_",
                string=rst
            )
        return rst

    def replace_autodoc_refs(self, rst: str) -> Tuple[str, Set]:
        """
        :param rst:
        :return:
        """
        # :role:`{~.}{module|class}{.}target` where {} is optional
        pattern = r":(?:mod|class|meth|func):`([~\.\w]+)`"

        if self.config.replace_attrs:
            if self.config.docs_url_type == 'html':
                # If linking to HTML docs, we can generate cross-refs for attributes
                pattern = r":(?:mod|class|meth|func|attr):`([~\.\w]+)`"
            else:
                # If linking to source code, just replace :attr:`~.attribute` with ``attribute``
                rst = self.replace_autodoc_attrs(rst)

        # To render on GitHub/PyPi/etc., we use Sphinx substitutions instead of cross-refs
        # Syntax is |.{ref}|_ or |.`{ref}`|_
        if self.config.inline_markup:
            repl = r"|.`\1`|_"
        else:
            repl = r"|.\1|_"

        # Get a list of all autodoc cross-refs
        autodoc_refs = set(re.findall(pattern, rst))

        # Replace cross-refs with substitutions
        rst = re.sub(pattern, repl, rst)

        # Use autodoc_refs and ref_map to generate substitution definitions
        return rst, autodoc_refs

    def replace_autodoc_attrs(self, rst: str) -> str:
        # Ex. ~.Class.meth => ``meth``
        short_ref = r" :attr:`~\.?(\w+)`"
        # Ex. .Class.meth => ``Class.meth``
        long_ref = r" :attr:`\.?([\.\w]+)`"
        repl = r" ``\1``"

        rst = re.sub(short_ref, repl, rst)
        rst = re.sub(long_ref, repl, rst)
        return rst

    def get_header_vals(self, autodoc_refs: Set) -> List[str]:
        header = []

        for ref in autodoc_refs:
            info = self.ref_map[ref]

            # Check for empty REFERENCE_MAPPING
            if not any(info.values()):
                continue

            if self.config.inline_markup:
                ref = f"`{ref}`"

            header.extend([
                f".. |.{ref}| replace:: {info['replace']}",
                f".. _.{ref}: {info['target']}"
            ])

        if not self.config.raw_directive:
            for _type, icon in self.config.icon_map.items():
                header.append(f'.. |{_type}| replace:: {icon}')

        return header

    def get_admonition_regex(self, admonition, admonition_type):
        body = escape_rst(admonition['body'])
        title = escape_rst(admonition['title'])

        # Account for arbitrary whitespace before each line of directive content
        lines = (line.replace('\n', '\n\s+') for line in body.split('\n\n'))
        body = '\n\n\s+'.join(lines)

        if admonition_type == 'specific':
            # For example, .. note:: This is a note
            pattern = fr"\.\. {admonition['class']}::\n*?\s+"

        else:
            # Generic admonition directives with/without class option
            pattern = rf"\.\. admonition::\s+{title}" + r"\n"

            if cls := admonition['class']:
                if 'admonition-' not in cls:
                    pattern += rf"\s+:class: {cls}" + r"\n"

            pattern += r"\n*?\s+"

        if not self.config.raw_directive:
            # csv-table template body uses match group
            pattern += rf"({body})"
        else:
            # raw html template body uses string formatting
            pattern += rf"{body}"

        pattern += r"(?=[\n\S]+?)"
        return pattern

    def get_admonition_icon(self, admonition: dict):
        icon = self.config.icon_map.get(admonition['class'])

        # Raw directive allows for using icon directly
        if self.config.raw_directive:
            return icon if icon else self.config.default_admonition_icon

        if icon:  # Without raw directive, must use substitution
            return f"|{admonition['class']}|"

        # Use default icon if admonition class isn't in icon map
        return "|default|"



import re
import subprocess
from pathlib import Path
from sphinx.application import Sphinx
from typing import Dict, List, Optional, Any, Union
from sphinx.util import logging


logger = logging.getLogger(__name__)


def get_conf_val(app: Sphinx, attr: str, default: Optional[Any] = None) -> Any:
    """Retrieve the value of a ``conf.py`` config variable

    :param attr: the config variable to retrieve
    :param default: the default value to return if the variable isn't found
    """
    return app.config._raw_config.get(attr, getattr(app.config, attr, default))


def set_conf_val(app: Sphinx, attr: str, value: Any) -> None:
    """Set the value of a ``conf.py`` config variable

    :param attr: the config variable to set
    :param value: the variable value
    """
    app.config._raw_config[attr] = value
    setattr(app.config, attr, value)


def escape_rst(rst: str) -> str:
    """Escape regex special characters from the content of an ``rst`` file"""
    for char in ".+?*|()<>{}^$[]":
        rst = rst.replace(char, rf"\{char}")
    return rst


def format_rst(inline_markup: str, rst: str) -> str:
    """Formats text with the specified type of inline markup

    Preserves any ``inline literals`` within the text

    :Example::

        >>> format_rst("bold", "This is part of the ``sphinx_readme.utils`` module")
        "**This is part of the** ``sphinx_readme.utils`` **module**"

    :param inline_markup: either "bold" or "italic"
    :param rst: the rst content to format
    """
    if inline_markup == "bold":
        markup = "**"
    elif inline_markup == "italic":
        markup = "*"
    else:
        raise ValueError("``inline_markup`` must be either 'bold' or 'italic'")

    split = re.split(r"(?<=\S)(\s*?``.+?``\s*?)(?=\S)", rst)
    parts = []

    for part in split:
        if "`" in part:
            parts.append(part.strip())
        else:
            parts.append(f"{markup}{part}{markup}")

    return " ".join(parts)


def replace_only_directives(rst: str) -> str:
    """Replaces and removes ``only`` directives

    If ``"readme"`` is in the ``<expression>`` part of the
    ``only`` directive, the content will be used.

    Otherwise, the directive will be removed

    :param rst: the content of an ``rst`` file
    """
    # Match all ``only`` directives
    pattern = r"\.\. only::\s+(\S.*?)\n+?((?:^[ ]+.+?$|^\s*$)+?)(?=\n*\S+|\Z)"
    directives = re.findall(pattern, rst, re.M | re.DOTALL)

    for expression, content in directives:
        # Pattern to match each block exactly
        pattern = rf"\.\. only:: {expression}\n+?{escape_rst(content)}\n*?"

        if 'readme' in expression:
            # For replacement, remove preceding indent (3 spaces) from each line
            content = '\n'.join(line[3:] for line in content.split('\n'))

            # Replace directive with content
            rst = re.sub(pattern, rf"{content}", rst)

        else:
            # Remove directive
            rst = re.sub(pattern, '', rst)

    return rst


def remove_raw_directives(rst: str) -> str:
    return re.sub(
        pattern=r"(\.\. raw::\s+\S.*?\n+?(?:^[ ]+.+?$|^\s*$)+?)(?=\n*\S+|\Z)",
        repl='', string=rst, flags=re.M | re.DOTALL
    )


def get_variants(obj: str):
    """

    >>> get_variants('mod.Class.meth')
    >>> ['mod.Class.meth', '.mod.Class.meth', '~mod.Class.meth', '~.mod.Class.meth']
    """
    return [prefix + obj for prefix in ('', '.', '~', '~.')]


def get_all_variants(fully_qualified_name: str) -> List[str]:
    """Generates a list of all possible ways to cross-reference a class/method/function

    >>> get_all_variants("sphinx_github_style.meth_lexer.TDKMethLexer.get_pkg_lexer")

    ['get_pkg_lexer', '.get_pkg_lexer', '~get_pkg_lexer', '~.get_pkg_lexer', 'TDKMethLexer.get_pkg_lexer',
    '.TDKMethLexer.get_pkg_lexer', '~TDKMethLexer.get_pkg_lexer', '~.TDKMethLexer.get_pkg_lexer',
    'meth_lexer.TDKMethLexer.get_pkg_lexer', '.meth_lexer.TDKMethLexer.get_pkg_lexer',
    '~meth_lexer.TDKMethLexer.get_pkg_lexer', '~.meth_lexer.TDKMethLexer.get_pkg_lexer',
    'sphinx_github_style.meth_lexer.TDKMethLexer.get_pkg_lexer',
     '.sphinx_github_style.meth_lexer.TDKMethLexer.get_pkg_lexer',
     '~sphinx_github_style.meth_lexer.TDKMethLexer.get_pkg_lexer',
      '~.sphinx_github_style.meth_lexer.TDKMethLexer.get_pkg_lexer']

    :param fully_qualified_name: the fully qualified name (pkg.module.class.method)
    """
    parts = fully_qualified_name.split(".")[::-1]  # => ['meth', 'Class', 'mod', "pkg"]
    variants = []

    for i, part in enumerate(parts):
        ref = '.'.join(parts[i::-1])  # 'meth', 'Class.meth', 'mod.class.meth', 'pkg.mod.class.meth'
        variants.extend(get_variants(ref))

    return variants
